==> canl-c%2B%2B-0.1.3rc4.tar.gz: canl-c++-0.1.3rc4/src/canl-c++/validator.cpp <==
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/stat.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/pkcs12.h>
#include <openssl/opensslv.h>
#include <openssl/bn.h>
#include <openssl/err.h>

#include "dcisec.h"
#include "logger.h"
#include "opensslutil.h"

#define OpenSSLError() \
{ \
        unsigned long errors; \
        char errors_buf[256] = { 0 }; \
        do{ \
                errors = ERR_get_error(); \
                if (errors != 0) { \
                        ValidatorLogger.msg(ERROR, "OpenSSL error string: %s", ERR_error_string(errors, errors_buf)); \
                } \
        }  while(errors != 0); \
}

namespace AuthN {

using namespace AuthN::OpenSSL;

#define X509_NAME_MAXLEN 256

  Logger& ValidatorLogger = log();

  Validator::Validator(const AuthN::Context& ctx) : mode_(Validator::ValidationNone) { context_ = &ctx.Copy(); }

  Validator::Validator(const AuthN::Context& ctx, ValidationMode mode) : mode_(mode) {
    context_ = &ctx.Copy();
  }

  Validator::~Validator(void) {
    if(context_) delete context_;
  }

  Validator::operator bool(void) const { return true; }

  bool Validator::operator!(void) const { return false; }

  Validator& Validator::Copy(void) const { 
    Validator* validator = new Validator(*this);
    validator->SetContext(*context_);
    return *validator;
  }

  Validator::ValidationMode Validator::Mode(void) const { return mode_; }

  void Validator::SetMode(ValidationMode mode) { mode_ = mode; }

  const AuthN::Context& Validator::GetContext(void) const { return *context_; }

  void Validator::SetContext(const AuthN::Context& env) { context_ = &env.Copy(); }

  //We need an explicit checking of the valitidy of crl here, since 
  //we need to get the validity for the validate mode option ValidationCRLIfValid
  static bool verify_crl(const std::string& crl_file, const std::string& CApath) {
    bool ret = false;
    X509_CRL* crl = NULL;
    X509_STORE* verify_store = NULL;
    X509_STORE_CTX ctx;
    X509_LOOKUP* lookup = NULL;
    X509_OBJECT obj;
    EVP_PKEY* pkey = NULL;
    int i;

    verify_store = setup_verify("", CApath);
    if(!verify_store) goto end;
    if(!X509_STORE_CTX_init(&ctx, verify_store, NULL, NULL)) {
      ValidatorLogger.msg(ERROR, "Failed to initialize X509 store");
      goto end;
    }
    loadCRLFile(crl_file, crl);
    if(!crl) goto end;
    i = X509_STORE_get_by_subject(&ctx, X509_LU_X509, X509_CRL_get_issuer(crl), &obj);
    if(i <= 0) {
      ValidatorLogger.msg(ERROR, "Failed to get CRL issuer certificate");
      goto end;
    }
    pkey = X509_get_pubkey(obj.data.x509);
    if(!pkey) {
      ValidatorLogger.msg(ERROR, "Failed to get CRL issuer public key");
      goto end;
    }
    i = X509_CRL_verify(crl, pkey);
    if(i < 0) goto end;
    if(i == 0) ValidatorLogger.msg(ERROR, "Failed to verify crl");
    else ValidatorLogger.msg(INFO, "Succeeded to verify crl");
    ret = true;

    end:
    X509_OBJECT_free_contents(&obj);
    if(pkey) EVP_PKEY_free(pkey);
    if(crl) X509_CRL_free(crl);
    if(verify_store) {
      X509_STORE_CTX_cleanup(&ctx);
      X509_STORE_free(verify_store);
    }
    return ret;
  }

  Status Validator::Validate(Credentials& cred) {
    Status status(-1);
    std::string capath;
    X509_STORE* cert_store = NULL;
    X509_STORE_CTX* store_ctx = NULL;
    cert_store = X509_STORE_new();
    X509_STORE_set_verify_cb_func(cert_store, verify_callback);

    X509* cert = NULL;
    X509* issuer = NULL;
    EVP_PKEY* key = NULL;
    STACK_OF(X509)* certchain = NULL; 
    //In case of proxy cert, the chain is supposed to be the certificates between proxy and CA cert.
    cert = cred.GetCertificate();
    certchain = cred.GetChain();
    if (certchain != NULL) {
      for (int i=0;i<sk_X509_num(certchain);i++) {
        X509* cert_in_chain = sk_X509_value(certchain,i);
        int j = X509_STORE_add_cert(cert_store, cert_in_chain);
        if (!j) {
          if ((ERR_GET_REASON(ERR_peek_error()) == X509_R_CERT_ALREADY_IN_HASH_TABLE)) {
            ERR_clear_error();
            break;
          }
          else { OpenSSLError(); goto err;}
        }
      }
      issuer = sk_X509_value(certchain, 0);
    }
    key = cred.GetPrivateKey();
 
    capath = context_->GetCAPath();
    if (X509_STORE_load_locations(cert_store, NULL,
           capath.empty() ? NULL:capath.c_str())) {
      store_ctx = X509_STORE_CTX_new();
      X509_STORE_CTX_init(store_ctx, cert_store, cert, certchain);

      /* override the check_issued with our version */
      store_ctx->check_issued = check_issued;
     
      //Handle the CRL 
      //Check if the corresponding CRL file is available
      bool crl_valid = false;
      bool crl_present = false;
      if(!(mode_&ValidationCRLNone)) {
        std::string hash_str;
        STACK_OF(X509)* chain = cred.GetChain();
        int n = sk_X509_num(chain);
        X509* tmp = NULL;
        if(chain && n) tmp = sk_X509_value(chain, n-1);
        char hash[32];
        memset(hash, 0, 32);
        snprintf(hash, 32, "%08lx", X509_subject_name_hash(tmp));
        hash_str = hash;
        std::string crl_loc;
        crl_loc = context_->GetCAPath() + "/" + hash_str + ".crl";
        struct stat st;
        if(stat(crl_loc.c_str(),&st) != 0) 
          if(mode_ & ValidationCRLMandatory) ValidatorLogger.msg(ERROR,"Failed to find the crl file under: %s", crl_loc.c_str());
        if(!S_ISREG(st.st_mode)) 
          if(mode_ & ValidationCRLMandatory) ValidatorLogger.msg(ERROR, "The location: %s is not a regular file", crl_loc.c_str());
        ValidatorLogger.msg(VERBOSE, "The location of crl file: %s ",crl_loc.c_str());
        X509_LOOKUP* lookup = NULL;
        lookup = X509_STORE_add_lookup(cert_store,X509_LOOKUP_file());
        if(!lookup || (!X509_load_crl_file(lookup,crl_loc.c_str(), X509_FILETYPE_PEM)) ) {
          if(mode_ & ValidationCRLMandatory)
            ValidatorLogger.msg(ERROR, "Failed to load crl file: %s", crl_loc.c_str());
        }
        else {
          crl_present = true;
          //Check if the crl is valid
          if(mode_ & ValidationCRLIfValid) //Only check validity if the mode is set
            if(verify_crl(crl_loc, capath)) crl_valid = true;
        }
      }
      bool check_crl = false;
      bool allow_proxy = true; // Allow proxy by default
      if((crl_valid && (mode_ & ValidationCRLIfValid)) ||
         (crl_present && (mode_ & ValidationCRLIfPresent)) ||
         (mode_ & ValidationCRLMandatory)) 
        check_crl = true;
      if(mode_ & ValidationProxyAllow) allow_proxy = true;
      else if(mode_ & ValidationProxyDisallow) allow_proxy = false;
      if(check_crl && allow_proxy) {
#ifdef HAVE_OPENSSL_PROXY
        X509_STORE_CTX_set_flags(store_ctx, X509_V_FLAG_CRL_CHECK | X509_V_FLAG_ALLOW_PROXY_CERTS);
#else
        ValidatorLogger.msg(ERROR, "The proxy certificate is not allowed by this OpenSSL with version: %0x", OPENSSL_VERSION_NUMBER);
        X509_STORE_CTX_set_flags(store_ctx, X509_V_FLAG_CRL_CHECK);
#endif
      }
      else if(!check_crl && allow_proxy){
#ifdef HAVE_OPENSSL_PROXY
        X509_STORE_CTX_set_flags(store_ctx, X509_V_FLAG_ALLOW_PROXY_CERTS);
#else
        ValidatorLogger.msg(ERROR, "The proxy certificate is not allowed by this OpenSSL with version: %0x", OPENSSL_VERSION_NUMBER);
#endif
      }
      else if(check_crl && !allow_proxy) {
        X509_STORE_CTX_set_flags(store_ctx, X509_V_FLAG_CRL_CHECK);
      }

      //Handle the OCSP
      //get the ocsp responder url from AIA (Authority Info Access, authorityInfoAccess) of EEC
      //alternatively, the responder url could be set as configuration info
#ifdef HAVE_OPENSSL_PROXY
      int pos1 = X509_get_ext_by_NID(cert, NID_proxyCertInfo, -1);
      if(pos1 < 0)
#endif
      {
        std::string aia_name = "authorityInfoAccess";
        Credentials::Extension e;
        bool check_ocsp = false;
        bool res;
        std::string ocsp_uri;
        res = cred.GetExtension(aia_name, e);
        std::string str = e.value;
        //authorityInfoAccess = OCSP;URI:http://ocsp.my.host/
        std::string::size_type find = str.find("OCSP");
        if(find != std::string::npos) {
          find = str.find("URI:", find+1);
          if(find != std::string::npos)
            ocsp_uri = str.substr(find+4); 
        }
        ValidatorLogger.msg(INFO, "The uri of the ocsp responder is: %s", ocsp_uri.c_str());
        if(mode_ & ValidationOCSPMandatory) { 
          if(!res) ValidatorLogger.msg(ERROR,"Failed to get AIA information from certificate extension");
          else if(ocsp_uri.empty()) ValidatorLogger.msg(ERROR, "Failed to get OCSP information from AIA");
          else check_ocsp = true;
        }
        else if(mode_ & ValidationOCSPIfPresent) {
          if(!res) ValidatorLogger.msg(VERBOSE,"Unable to get AIA information from certificate extension");
          else if(ocsp_uri.empty()) ValidatorLogger.msg(VERBOSE, "Unable to get OCSP information from AIA");
          else check_ocsp = true;
        }
        if(check_ocsp) {
          bool res;
          std::string resp_str;
          std::string cert_path = context_->GetCertPath();
          res = AuthN::query_ocsp_responder(ocsp_uri, cert, issuer, cert, key, capath, "", cert_path, resp_str);
          if(!res) ValidatorLogger.msg(INFO, "The certificate is revoked");
        }
      }

      ValidationContext vctx;
      vctx.ca_path = context_->GetCAPath(); 
      if (!X509_STORE_CTX_set_ex_data(store_ctx, get_proxy_auth_ex_data_idx(), &vctx)) { 
        ValidatorLogger.msg(ERROR,"Can not set the STORE_CTX for chain verification");
        goto err;
      }
      //X509_STORE_CTX_set_depth(store_ctx, 10);
      if(!X509_verify_cert(store_ctx)) { goto err; }

      ProxyPolicyList* ppl = &(vctx.ppl);
      Credentials::Extension* ext = new Credentials::Extension;

      int pos = -1;
#ifdef HAVE_OPENSSL_PROXY
      pos = X509_get_ext_by_NID(cert, NID_proxyCertInfo, -1);
#endif
      if(pos >= 0)
      {
        X509_EXTENSION* ex = X509_get_ext(cert, pos);
        if(ex) {
          PROXY_CERT_INFO_EXTENSION *pci = (PROXY_CERT_INFO_EXTENSION *)X509V3_EXT_d2i(ex);
          std::string oid;
          char buf[X509_NAME_MAXLEN];
          int buflen;
          buflen = OBJ_obj2txt(buf, sizeof(buf), pci->proxyPolicy->policyLanguage, 0);
          if(buflen < 0) { ValidatorLogger.msg(ERROR,"Failed to call OBJ_obj2txt to convert obj to txt");}
          else oid = buf;
          ext->oid = oid;
          if(X509_EXTENSION_get_critical(ex)) ext->critical = true;
        }
        std::string policy_str;
        for(std::list<ProxyPolicy>::iterator it = (*ppl).begin(); it!= (*ppl).end(); it++) {
          policy_str.append("------BEGIN PROXY POLICY------").append("\n").append((*it).subject_name).append("\n").append((*it).policy).append("------END PROXY POLICY------").append("\n");
        }
        ext->value = policy_str;
        cred.proxy_policy_ = ext;
      }
    }
    //Replace the trusted certificate chain after verification passed, the
    //trusted ca certificate is added
    if(certchain) { sk_X509_pop_free(certchain, X509_free); }
    certchain = sk_X509_new_null();

    if(store_ctx != NULL) for (int i=0; i < sk_X509_num(store_ctx->chain); i++) {
      X509* tmp = NULL; tmp = X509_dup(sk_X509_value(store_ctx->chain,i));
      sk_X509_insert(certchain, tmp, i);
    }

    status = Status(0);

err:
    if(cert_store) { X509_STORE_free(cert_store); }
    if(store_ctx) { X509_STORE_CTX_free(store_ctx); }

    return status;
  }

}

==> canl-c%2B%2B-0.2.0.tar.gz: canl-c++-0.2.0/src/canl-c++/validator.cpp <==
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/stat.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/pkcs12.h>
#include <openssl/opensslv.h>
#include <openssl/bn.h>
#include <openssl/err.h>

#include "dcisec.h"
#include "logger.h"
#include "opensslutil.h"

#define OpenSSLError() \
{ \
        unsigned long errors; \
        char errors_buf[256] = { 0 }; \
        do{ \
                errors = ERR_get_error(); \
                if (errors != 0) { \
                        ValidatorLogger.msg(ERROR, "OpenSSL error string: %s", ERR_error_string(errors, errors_buf)); \
                } \
        }  while(errors != 0); \
}

namespace AuthN {

using namespace AuthN::OpenSSL;

#define X509_NAME_MAXLEN 256

  Logger& ValidatorLogger = log();

  Validator::Validator(const AuthN::Context& ctx) : mode_(Validator::ValidationNone) { context_ = &ctx.Copy(); }

  Validator::Validator(const AuthN::Context& ctx, ValidationMode mode) : mode_(mode) {
    context_ = &ctx.Copy();
  }

  Validator::~Validator(void) {
    if(context_) delete context_;
  }

  Validator::operator bool(void) const { return true; }

  bool Validator::operator!(void) const { return false; }

  Validator& Validator::Copy(void) const { 
    Validator* validator = new Validator(*this);
    validator->SetContext(*context_);
    return *validator;
  }

  Validator::ValidationMode Validator::Mode(void) const { return mode_; }

  void Validator::SetMode(ValidationMode mode) { mode_ = mode; }

  const AuthN::Context& Validator::GetContext(void) const { return *context_; }

  void Validator::SetContext(const AuthN::Context& env) { context_ = &env.Copy(); }

  // Check CRL times
  static bool check_crl_time(X509_CRL* crl) {
    time_t now;
    int i;
    i=X509_cmp_time(X509_CRL_get_lastUpdate(crl), &now);
    if(i == 0) {
      ValidatorLogger.msg(ERROR, "Error in CRL last updated field");
      return false;
    }
    if(i > 0) {
      ValidatorLogger.msg(ERROR, "CRL is not yet valid");
      return false;
    }
    if(X509_CRL_get_nextUpdate(crl)) {
      i = X509_cmp_time(X509_CRL_get_nextUpdate(crl), &now);
      if(i == 0) {
        ValidatorLogger.msg(ERROR, "Error in CRL next updated field");
        return false;
      }
      if(i < 0) {
        ValidatorLogger.msg(ERROR, "CRL has expired"); 
        return false;
      }
    }
    return true;
  }

  //We need an explicit checking of the valitidy of crl here, since 
  //we need to get the validity for the validate mode option ValidationCRLIfValid
  //This is different with the openssl method X509_verify_cert, since it check the validity
  //of crl file when the certificate is checked. We here need to independently check 
  //the validity of crl before the certificate is checked.
  static bool verify_crl(const std::string& crl_file, const std::string& CApath, bool check_time) {
    bool ret = false;
    X509_CRL* crl = NULL;
    X509_STORE* verify_store = NULL;
    X509_STORE_CTX ctx;
    X509_LOOKUP* lookup = NULL;
    X509_OBJECT obj;
    EVP_PKEY* pkey = NULL;
    int i;

    verify_store = setup_verify("", CApath);
    if(!verify_store) goto end;
    if(!X509_STORE_CTX_init(&ctx, verify_store, NULL, NULL)) {
      ValidatorLogger.msg(ERROR, "Failed to initialize X509 store");
      goto end;
    }
    loadCRLFile(crl_file, crl);
    if(!crl) goto end;

    //Check the validity of lifetime
    
    if(check_time && !check_crl_time(crl)) {
      ValidatorLogger.msg(ERROR, "Failed to check the validity of lifetime");
      return false;
    }

    //Check the validity of signature
    i = X509_STORE_get_by_subject(&ctx, X509_LU_X509, X509_CRL_get_issuer(crl), &obj);
    if(i <= 0) {
      ValidatorLogger.msg(ERROR, "Failed to get CRL issuer certificate");
      goto end;
    }
    pkey = X509_get_pubkey(obj.data.x509);
    if(!pkey) {
      ValidatorLogger.msg(ERROR, "Failed to get CRL issuer public key");
      goto end;
    }
    i = X509_CRL_verify(crl, pkey);
    if(i < 0) goto end;
    if(i == 0) ValidatorLogger.msg(ERROR, "Failed to verify crl");
    else ValidatorLogger.msg(INFO, "Succeeded to verify crl");
    ret = true;

    end:
    X509_OBJECT_free_contents(&obj);
    if(pkey) EVP_PKEY_free(pkey);
    if(crl) X509_CRL_free(crl);
    if(verify_store) {
      X509_STORE_CTX_cleanup(&ctx);
      X509_STORE_free(verify_store);
    }
    return ret;
  }

  static std::vector<std::string> verify_crls(std::vector<std::string>& crl_files, const std::string& CApath, bool check_time) {
    //if check_time is true, then any crl file with invalid lifetime  will be excluded in the valid_crl_files
    std::vector<std::string> valid_crl_files;
    std::vector<std::string>::size_type size = crl_files.size();
    unsigned int i;
    for(i = 0; i <= size; i++) {
      struct stat st;
      if(stat(crl_files[i].c_str(),&st) != 0) {
        ValidatorLogger.msg(ERROR,"Failed to find the crl file under: %s", crl_files[i].c_str()); continue;
      }
      if(!S_ISREG(st.st_mode)) {
        ValidatorLogger.msg(ERROR, "The location: %s is not a regular file", crl_files[i].c_str()); continue;
      }

      int ret = verify_crl(crl_files[i], CApath, check_time);
      if(!ret && !check_time) {
        ValidatorLogger.msg(ERROR, "The crl file %s is not valid due to the failure of lifetime checking", crl_files[i].c_str());
      }
      else if(!ret && check_time) { 
        ValidatorLogger.msg(ERROR, "The crl file %s is not valid due to the failure of signature checking", crl_files[i].c_str());
      }
      else { valid_crl_files.push_back(crl_files[i]); }
    }
    return valid_crl_files;
  }

  Status Validator::Validate(Credentials& cred) {
    Status status(-1);
    std::string capath;
    X509_STORE* cert_store = NULL;
    X509_STORE_CTX* store_ctx = NULL;
    cert_store = X509_STORE_new();
    X509_STORE_set_verify_cb_func(cert_store, verify_callback);

    X509* cert = NULL;
    X509* issuer = NULL;
    EVP_PKEY* key = NULL;
    STACK_OF(X509)* certchain = NULL; 
    //In case of proxy cert, the chain is supposed to be the certificates between proxy and CA cert.
    cert = cred.GetCertificate();
    certchain = cred.GetChain();
    if (certchain != NULL) {
      for (int i=0;i<sk_X509_num(certchain);i++) {
        X509* cert_in_chain = sk_X509_value(certchain,i);
        int j = X509_STORE_add_cert(cert_store, cert_in_chain);
        if (!j) {
          if ((ERR_GET_REASON(ERR_peek_error()) == X509_R_CERT_ALREADY_IN_HASH_TABLE)) {
            ERR_clear_error();
            break;
          }
          else { OpenSSLError(); goto err;}
        }
      }
      issuer = sk_X509_value(certchain, 0);
    }
    key = cred.GetPrivateKey();
 
    capath = context_->GetCAPath();
    if (X509_STORE_load_locations(cert_store, NULL,
           capath.empty() ? NULL:capath.c_str())) {
      store_ctx = X509_STORE_CTX_new();
      X509_STORE_CTX_init(store_ctx, cert_store, cert, certchain);

      /* override the check_issued with our version */
      store_ctx->check_issued = check_issued;
     
      //Handle the CRL 
      //Check if the corresponding CRL file is available
      bool crl_valid = false;
      bool crl_present = false;
      if(!(mode_&ValidationCRLNone)) {
        std::string hash_str;
        STACK_OF(X509)* chain = cred.GetChain();
        int n = sk_X509_num(chain);
        X509* tmp = NULL;
        if(chain && n) tmp = sk_X509_value(chain, n-1);

        X509_LOOKUP* lookup = NULL;
        lookup = X509_STORE_add_lookup(cert_store,X509_LOOKUP_file());

        //Check the valitidy of crl files, including the crl files that is added,
        //and the default crl file under CA directory
       
        //Check crl files that are added
        std::vector<std::string> crl_files = context_->GetCRLFiles();     
        std::vector<std::string> verified_crl_files;

        //The default crl under CA directory
        char hash[32];
        memset(hash, 0, 32);
        snprintf(hash, 32, "%08lx", X509_subject_name_hash(tmp));
        hash_str = hash;
        std::string crl_loc;
        crl_loc = context_->GetCAPath() + "/" + hash_str + ".crl";
        struct stat st;
        if(stat(crl_loc.c_str(),&st) != 0) {
          if((crl_files.size() == 0) && (mode_ & ValidationCRLMandatory)) {
            ValidatorLogger.msg(ERROR,"Failed to find the crl file under: %s", crl_loc.c_str());
            goto err;
          }
        }
        else {
          if(!S_ISREG(st.st_mode)) {
            if((crl_files.size() ==0) && (mode_ & ValidationCRLMandatory)) {
              ValidatorLogger.msg(ERROR, "The location: %s is not a regular file", crl_loc.c_str());
              goto err;
            }
          }
        }
        crl_files.push_back(crl_loc);

        //Check the validity of crl file
        if(mode_ & ValidationCRLIfPresent)
          verified_crl_files = verify_crls(crl_files, capath, false);
        if(mode_ & ValidationCRLIfValid)
          verified_crl_files = verify_crls(crl_files, capath, true);
        else {
          //std::vector<std::string>::size_type size = crl_files.size();
          verified_crl_files.insert(verified_crl_files.begin(), crl_files.begin(), crl_files.end());
          //unsigned int i = crl_files
        }

        if(lookup != NULL) {
          std::vector<std::string>::size_type size = verified_crl_files.size();
          for(unsigned int i = 0; i<size; i++) {
            int res = X509_load_crl_file(lookup, verified_crl_files[i].c_str(), X509_FILETYPE_PEM);
            if(!res) res = X509_load_crl_file(lookup, verified_crl_files[i].c_str(), X509_FILETYPE_ASN1);
            if(!res) {
              if(mode_ & ValidationCRLMandatory)
                ValidatorLogger.msg(ERROR, "Failed to load crl file: %s", verified_crl_files[i].c_str());
            }
            else {
              crl_present = true;
            }
          }
        }
      }

      bool check_crl = false;
      bool allow_proxy = true; // Allow proxy by default
      if((crl_valid && (mode_ & ValidationCRLIfValid)) ||
         (crl_present && (mode_ & ValidationCRLIfPresent)) ||
         (mode_ & ValidationCRLMandatory)) 
        check_crl = true;
      if(mode_ & ValidationProxyAllow) allow_proxy = true;
      else if(mode_ & ValidationProxyDisallow) allow_proxy = false;
      if(check_crl && allow_proxy) {
#ifdef HAVE_OPENSSL_PROXY
        X509_STORE_CTX_set_flags(store_ctx, X509_V_FLAG_CRL_CHECK | X509_V_FLAG_ALLOW_PROXY_CERTS);
#else
        ValidatorLogger.msg(ERROR, "The proxy certificate is not allowed by this OpenSSL with version: %0x", OPENSSL_VERSION_NUMBER);
        X509_STORE_CTX_set_flags(store_ctx, X509_V_FLAG_CRL_CHECK);
#endif
      }
      else if(!check_crl && allow_proxy){
#ifdef HAVE_OPENSSL_PROXY
        X509_STORE_CTX_set_flags(store_ctx, X509_V_FLAG_ALLOW_PROXY_CERTS);
#else
        ValidatorLogger.msg(ERROR, "The proxy certificate is not allowed by this OpenSSL with version: %0x", OPENSSL_VERSION_NUMBER);
#endif
      }
      else if(check_crl && !allow_proxy) {
        X509_STORE_CTX_set_flags(store_ctx, X509_V_FLAG_CRL_CHECK);
      }

      //Handle the OCSP
      //get the ocsp responder url from AIA (Authority Info Access, authorityInfoAccess) of EEC
      //alternatively, the responder url could be set as configuration info
#ifdef HAVE_OPENSSL_PROXY
      int pos1 = X509_get_ext_by_NID(cert, NID_proxyCertInfo, -1);
      if(pos1 < 0)
#endif
      {
        std::string aia_name = "authorityInfoAccess";
        Credentials::Extension e;
        bool check_ocsp = false;
        bool res;
        std::string ocsp_uri;
        res = cred.GetExtension(aia_name, e);
        std::string str = e.value;
        //authorityInfoAccess = OCSP;URI:http://ocsp.my.host/
        std::string::size_type find = str.find("OCSP");
        if(find != std::string::npos) {
          find = str.find("URI:", find+1);
          if(find != std::string::npos)
            ocsp_uri = str.substr(find+4); 
        }
        ValidatorLogger.msg(INFO, "The uri of the ocsp responder is: %s", ocsp_uri.c_str());
        if(mode_ & ValidationOCSPMandatory) { 
          if(!res) ValidatorLogger.msg(ERROR,"Failed to get AIA information from certificate extension");
          else if(ocsp_uri.empty()) ValidatorLogger.msg(ERROR, "Failed to get OCSP information from AIA");
          else check_ocsp = true;
        }
        else if(mode_ & ValidationOCSPIfPresent) {
          if(!res) ValidatorLogger.msg(VERBOSE,"Unable to get AIA information from certificate extension");
          else if(ocsp_uri.empty()) ValidatorLogger.msg(VERBOSE, "Unable to get OCSP information from AIA");
          else check_ocsp = true;
        }
        if(check_ocsp) {
          bool res;
          std::string resp_str;
          std::string cert_path = context_->GetCertPath();
          res = AuthN::query_ocsp_responder(ocsp_uri, cert, issuer, cert, key, capath, "", cert_path, resp_str);
          if(!res) ValidatorLogger.msg(INFO, "The certificate is revoked");
        }
      }

      ValidationContext vctx;
      vctx.ca_path = context_->GetCAPath(); 
      if (!X509_STORE_CTX_set_ex_data(store_ctx, get_proxy_auth_ex_data_idx(), &vctx)) { 
        ValidatorLogger.msg(ERROR,"Can not set the STORE_CTX for chain verification");
        goto err;
      }
      //X509_STORE_CTX_set_depth(store_ctx, 10);
      if(!X509_verify_cert(store_ctx)) { goto err; }

      ProxyPolicyList* ppl = &(vctx.ppl);
      Credentials::Extension* ext = new Credentials::Extension;

      int pos = -1;
#ifdef HAVE_OPENSSL_PROXY
      pos = X509_get_ext_by_NID(cert, NID_proxyCertInfo, -1);
#endif
      if(pos >= 0)
      {
        X509_EXTENSION* ex = X509_get_ext(cert, pos);
        if(ex) {
          PROXY_CERT_INFO_EXTENSION *pci = (PROXY_CERT_INFO_EXTENSION *)X509V3_EXT_d2i(ex);
          std::string oid;
          char buf[X509_NAME_MAXLEN];
          int buflen;
          buflen = OBJ_obj2txt(buf, sizeof(buf), pci->proxyPolicy->policyLanguage, 0);
          if(buflen < 0) { ValidatorLogger.msg(ERROR,"Failed to call OBJ_obj2txt to convert obj to txt");}
          else oid = buf;
          ext->oid = oid;
          if(X509_EXTENSION_get_critical(ex)) ext->critical = true;
        }
        std::string policy_str;
        for(std::list<ProxyPolicy>::iterator it = (*ppl).begin(); it!= (*ppl).end(); it++) {
          policy_str.append("------BEGIN PROXY POLICY------").append("\n").append((*it).subject_name).append("\n").append((*it).policy).append("------END PROXY POLICY------").append("\n");
        }
        ext->value = policy_str;
        cred.proxy_policy_ = ext;
      }
    }
    //Replace the trusted certificate chain after verification passed, the
    //trusted ca certificate is added
    if(certchain) { sk_X509_pop_free(certchain, X509_free); }
    certchain = sk_X509_new_null();

    if(store_ctx != NULL) for (int i=0; i < sk_X509_num(store_ctx->chain); i++) {
      X509* tmp = NULL; tmp = X509_dup(sk_X509_value(store_ctx->chain,i));
      sk_X509_insert(certchain, tmp, i);
    }

    status = Status(0);

err:
    if(cert_store) { X509_STORE_free(cert_store); }
    if(store_ctx) { X509_STORE_CTX_free(store_ctx); }

    return status;
  }

}
